
#ifndef _GFT_IFT_H_
#define _GFT_IFT_H_

#include "gft_common.h"
#include "gft_imagegraph.h"
#include "gft_pqueue32.h"
#include "gft_queue.h"
#include "gft_heap.h"
#include "gft_stack.h"
#include "gft_graph.h"
#include "gft_set.h"
#include "gft_morphology.h"
#include "gft_analysis.h"
#include "gft_image32.h"
#include "gft_adjrel.h"

//----------------------------------------
//  Method wrappers:

//The 'label' scene should be pre-initialized as follows:
//  label->data[p]=NIL, unlabeled voxel.
//  label->data[p]=0,   background voxel.
//  label->data[p]=1,   object#1 voxel.
//  label->data[p]=2,   object#2 voxel.
//  ...
//---------------------------------------

namespace gft{
  namespace ift{


    //---------------------------------------
    /* Star Convexity - ICIP 2013 */
    sImage32 *SC_Pred_fsum(sImageGraph *sg,
			   int *S, float power);

    void SC_IFT(sImageGraph *sg,
		int *S,
		sImage32 *label,
		sImage32 *P_sum);
    //---------------------------------------
    
    //Outer Cut:
    int GetEnergy_Min(sImageGraph *sg,
		      sImage32 *label,
		      int lb);
    int GetEnergy_Max(sImageGraph *sg,
		      sImage32 *label,
		      int lb);
    long long GetEnergy_Sum(sImageGraph *sg,
			    sImage32 *label,
			    int lb);

    int GetEnergy_Min(sGraph *graph,
		      int *label,
		      int lb);
    
    //---------------------------------------
    // ORFC:

    //Outer Cut (the polarity must be embedded in the graph):
    void ORFC(sImageGraph *sg,
	      int *S,
	      sImage32 *label);
 

    //---------------------------------------
    // OIFT:
    
    //Outer Cut:
    void OIFT(sImage32 *W,
	      sAdjRel *A,
	      sImage32 *img,
	      float per,
	      int *S,
	      sImage32 *label);

    
    //Outer Cut (the polarity must be embedded in the graph):
    void OIFT(sImageGraph *sg,
	      int *S,
	      sImage32 *label);

    void OIFT_MinMax(sImageGraph *sg,
		     int *S,
		     sImage32 *label,
		     sImage32 *pred,
		     sImage32 *value,
		     int niter);

    void OIFT_MaxMin(sImageGraph *sg,
		     int *S,
		     sImage32 *label,
		     sImage32 *pred,
		     sImage32 *value,
		     int niter);
    
    void OIFT_TZ2Bkg(sImageGraph *sg,
		     int *S,
		     sImage32 *label);
    void OIFT_TZ2Obj(sImageGraph *sg,
		     int *S,
		     sImage32 *label);
    
    void OIFT_TZ(sImageGraph *sg,
		 int *S,
		 sImage32 *label);

    bool isOIFT(sImageGraph *sg,
		int *S,
		sImage32 *Slabel,
		sImage32 *label);
    
    //Outer Cut (the polarity must be embedded in the graph):
    void OIFT(sGraph *graph,
	      sGraph *transpose,
	      int *S,
	      int *label);

    //Inner Cut (the polarity must be embedded in the graph):
    void OIFT_in(sImageGraph *sg,
		 int *S,
		 sImage32 *label);

    void OIFT_Heap(sImageGraph *sg,
		   int *S,
		   sImage32 *label);
    void OIFT_Heap(sGraph *graph,
		   sGraph *transpose,
		   int *S,
		   int *label);
        
    //Outer Cut (the polarity must be embedded in the graph):
    //Energy based implementations.
    void EOIFT(sImageGraph *sg,
	       int *S,
	       sImage32 *label);
    void EOIFT_2(sImageGraph *sg,
		 int *S,
		 sImage32 *label);    

    void EOIFT(sGraph *graph,
	       sGraph *transpose,
	       int *S,
	       int *label);
    
    void EOIFT_Heap(sImageGraph *sg,
		    int *S,
		    sImage32 *label);
    void EOIFT_Heap_2(sImageGraph *sg,
		      int *S,
		      sImage32 *label);
    //Pode apagar depois a versao 3
    int EOIFT_Heap_3(sImageGraph *sg,
		     int *S,
		     sImage32 *label);

    void EOIFT_Heap(sGraph *graph,
		    sGraph *transpose,
		    int *S,
		    int *label);

    //---------------------------------------

    void IFT_fmax(sGraph *graph,
		  int *S,
		  int *label,
		  int *cost);
     
    void IFT_fmax_Heap(sGraph *graph,
		       int *S,
		       int *label,
		       float *cost);

    void IFT_fw(sImageGraph *g,
		int *S,
		sImage32 *label,
		sImage32 *cost,
		sImage32 *pred);
    
    void IFT_fw(sGraph *graph,
		int *S,
		int *label,
		int *cost);

    void IFT_fw(sGraph *graph,
		int *S,
		int *label,
		int *cost,
		int *pred);
        
    void IFT_fw_Heap(sGraph *graph,
		     int *S,
		     int *label,
		     float *cost,
		     int *pred);
    
    //---------------------------------------
    sImage32 *Cost_fmin(sImageGraph *sg,
			int *S, int lb,
			sImage32 *label);

    //---------------------------------------
    // OIFT with Connectivity Constraints:

    /* To be used by COIFT */
    sImage32 *Cost_fmax(sImageGraph *sg,
			int *S, int lb,
			sImage32 *label);
    
    void COIFT(sImageGraph *sg,
	       int *S,
	       sImage32 *label);
  
    //---------------------------------------
    //Boundary Band (BB) - ICIP 2014

    sImage32 *BB_Geodesic_Cost(sImage32 *pred,
			       sAdjRel *A);

    sImage32 *BB_CropTemplate(sImage32 *cost_template,
			      int *S,
			      sImage32 *label,
			      int lb);
    
    void BB_OIFT(sImageGraph *sg,
		 int *S,
		 sImage32 *L,
		 sImage32 *C,
		 sImage32 *P,
		 float delta);


    void RBB_OIFT(sImageGraph *sg,
		  int *S,
		  sImage32 *L,
		  sImage32 *C,
		  sImage32 *P,
		  float delta);

    //---------------------------------------

    int *GetSeedsByLabel(int *S,
			 sImage32 *label,
			 int lb);
    int *GetAllInternalSeedsByLabel(int *S,
				    sImage32 *label,
				    int lb,
				    int *hr,
				    int nlayers);
    sImageGraph *GetPolarityGraph(sImageGraph *graph,
				  sCImage *cimg,
				  sImage32 *img,
				  int pol);
    
    //---------------------------------------

    //Compute watershed by fpeak from markers
    void IFT_fpeak(sImage32 *grad,
		   sAdjRel *A,
		   sImage32 *label);
    
    //Compute watershed by fwv from markers
    void IFT_fwv(sImage32 *grad,
		 sAdjRel *A,
		 sImage32 *label);

    //--------------------------------------

    void RelaxMobj(sImageGraph *sg,
		   int *S,
		   sImage32 *label,
		   int ntimes);

    //beta = 1;
    float *ORelax_dual(sImageGraph *sg,
		       int *S,
		       sImage32 *label, int ntimes);

    //beta = 1;
    float *Relax_dual(sImageGraph *sg,
		      int *S,
		      sImage32 *label,
		      int ntimes);
    
    void Relax(sImageGraph *sg,
	       int *S,
	       sImage32 *label,
	       int ntimes);

    void ORelax_1(sImageGraph *sg,
		  int *S,
		  sImage32 *label,
		  int ntimes);

    void ORelax(sImageGraph *sg,
		int *S,
		sImage32 *label,
		int ntimes);

    void ORelax_i(sImageGraph *sg,
		  int *S,
		  sImage32 *label,
		  int ntimes);

    void ORelax_s(sImageGraph *sg,
		  int *S,
		  sImage32 *label,
		  int ntimes);    

    //--------------------------------------
    void GGC_maxmin(sImageGraph *graph,
		    sCImage *cimg,
		    sImage32 *img,
		    int method,
		    float power_geodesic,
		    float delta,
		    float theta_hedgehog,
		    float R,
		    int postproc,
		    int niterations,
		    int conn,
		    int pol,
		    int shapepriors,
		    int costtemplate,
		    int *S,
		    sImage32 *label);

    
    
  } //end ift namespace
} //end gft namespace

#endif

