
#include "gft.h"
#include "gft_connseeds.h"


gft::sCImage *ReadAnyCImage(char *file){
  gft::sCImage *cimg;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "ppm") == 0){
    cimg = gft::CImage::Read(file);
  }
  else{
    sprintf(command, "convert %s cimage_tmp.ppm", file);
    system(command);
    cimg = gft::CImage::Read("cimage_tmp.ppm");
    system("rm cimage_tmp.ppm");
  }
  return cimg;
}


gft::sImage32 *ReadAnyImage(char *file){
  gft::sImage32 *img;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "pgm") == 0){
    img = gft::Image32::Read(file);
  }
  else{
    sprintf(command, "convert %s image_tmp.pgm", file);
    system(command);
    img = gft::Image32::Read((char *)"image_tmp.pgm");
    system("rm image_tmp.pgm");
  }
  return img;
}



int *LoadConnInternalSeeds(int *P4){
  gft::sCImage *cimg;
  int *S;
  int p,i,n = 0;
  cimg = gft::CImage::Read((char *)"tree_min.ppm");
  for(i = 1; i <= P4[0]; i++){
    p = P4[i];
    cimg->C[0]->data[p] = 0;
    cimg->C[1]->data[p] = 0;
    cimg->C[2]->data[p] = 0;
  }

  for(p = 0; p < cimg->C[0]->n; p++){
    if(cimg->C[0]->data[p] != cimg->C[1]->data[p] ||
       cimg->C[0]->data[p] != cimg->C[2]->data[p])
      n++;
  }
  S = (int *)malloc((n+1)*sizeof(int));
  S[0] = n;
  i = 1;
  for(p = 0; p < cimg->C[0]->n; p++){
    if(cimg->C[0]->data[p] != cimg->C[1]->data[p] ||
       cimg->C[0]->data[p] != cimg->C[2]->data[p]){
      S[i] = p;
      i++;
    }
  }
  gft::CImage::Destroy(&cimg);
  return S;
}



void DrawSeeds(int *S, gft::sImage32 *img, int val){
  int i,p;
  for(i = 1; i <= S[0]; i++){
    p = S[i];
    img->data[p] = val;
  }
}


void WriteSeeds(int *Si, int *Se, gft::sImage32 *prob){
  gft::sCImage *cimg;
  int p,i;
  cimg = gft::CImage::Clone(prob);
  DrawSeeds(Si, cimg->C[0], 255);
  DrawSeeds(Si, cimg->C[1], 0);
  DrawSeeds(Si, cimg->C[2], 0);
  DrawSeeds(Se, cimg->C[0], 0);
  DrawSeeds(Se, cimg->C[1], 0);
  DrawSeeds(Se, cimg->C[2], 255);
  gft::CImage::Write(cimg, (char*)"seeds.ppm");
  gft::CImage::Destroy(&cimg);
}



int *LoadExternalSeeds(gft::sImage32 *dextrseg, int *P4,
		       int padding, int bordersize){
  ROI roi;
  int *S;
  int size,x,y,p,k,n;
  GetROI(&roi, P4, dextrseg, padding);
  n = dextrseg->n;
  size = n - (roi.x_max - roi.x_min + 1)*(roi.y_max - roi.y_min + 1);
  S = (int *)malloc((size+1)*sizeof(int));
  S[0] = size;
  k = 1;
  for(p = 0; p < n; p++){
    x = p % dextrseg->ncols;
    y = p / dextrseg->ncols;
    if(x > roi.x_max || x < roi.x_min ||
       y > roi.y_max || y < roi.y_min){
      if(x >= roi.x_min - bordersize && x <= roi.x_max + bordersize &&
	 y >= roi.y_min - bordersize && y <= roi.y_max + bordersize){
	S[k] = p;
	k++;
      }
    }
  }
  S[0] = k-1;
  S = (int *)realloc(S, k*sizeof(int));
  return S;
}



int *MergeSeeds(int *Si, int *Se){
  int *S;
  int i,j;
  S = (int *)malloc((Si[0]+Se[0]+1)*sizeof(int));
  S[0] = Si[0]+Se[0];
  j = 1;
  for(i = 1; i <= Si[0]; i++){
    S[j] = Si[i];
    j++;
  }
  for(i = 1; i <= Se[0]; i++){
    S[j] = Se[i];
    j++;
  }
  return S;
}



void WriteEDT(gft::sImage32 *edt){
  gft::sImage32 *tmp;
  int p;
  tmp = gft::Image32::Create(edt);
  for(p = 0; p < edt->n; p++){
    tmp->data[p] = ROUND(sqrtf(edt->data[p]));
  }
  gft::Image32::Write(tmp, (char *)"edt.pgm");
  gft::Image32::Destroy(&tmp);
}


gft::sImage32 *ComputeEnergyMap(gft::sCImage32f *cimg_lab,
				gft::sImage32 *prob,
				gft::sImage32 *dextrseg,
				int *Se,
				float w1,
				float w2){
  gft::sImage32 *edt,*edt8,*energy;
  gft::sImage32f *dist;
  gft::sAdjRel *A;
  int p, EDTmax;
  A = gft::AdjRel::Circular(1.5);
  edt = gft::Image32::Mask2EDT(dextrseg, A,
			       INTERIOR, INT_MAX, 0);
  //---------------------------------------
  //WriteEDT(edt);
  edt8 = gft::Image32::ConvertToNbits(edt, 8);
  dist = gft::Features::KNNMeanDistanceMap(cimg_lab, Se, 5, 100);
  energy = gft::Image32::Create(prob);
  for(p = 0; p < edt->n; p++)
    energy->data[p] = ROUND(w1*prob->data[p] + w2*edt8->data[p] + (1.0-w1-w2)*dist->data[p]);

  gft::Image32f::Destroy(&dist);
  gft::Image32::Destroy(&edt8);
  gft::Image32::Destroy(&edt);
  gft::AdjRel::Destroy(&A);
  return energy;
}


gft::sImageGraph *ComputeGraph(gft::sCImage32f *cimg_lab,
			       gft::sImage32 *membship,
			       float sharpness){
  gft::sImageGraph *G,*Gobj,*Gimg;
  gft::sImage32f *gradf;
  gft::sImage32 *grad;
  gradf = gft::CImage32f::SobelFilter(cimg_lab);
  grad = gft::Image32::Clone(gradf, 10.0);
  Gimg = gft::ImageGraph::ByWeightImage(grad, 1.5);
  gft::Image32f::Destroy(&gradf);
  gft::Image32::Destroy(&grad);

  grad = gft::Image32::SobelFilter(membship);
  Gobj = gft::ImageGraph::ByWeightImage(grad, 1.5);
  gft::ImageGraph::LinearStretch(Gobj,
				 0, Gobj->Wmax,
				 0, Gimg->Wmax);
  gft::Image32::Destroy(&grad);
  G = gft::ImageGraph::WeightedMean(Gimg, Gobj, sharpness);
  
  gft::ImageGraph::Destroy(&Gobj);
  gft::ImageGraph::Destroy(&Gimg);
  return G;
}




gft::sImage32 *SetInitialSeg(gft::sImage32 *dextrseg,
			     gft::sImage32 *membship,
			     float w1,
			     int *Si,
			     int *Se){
  gft::sImage32 *Seg;
  int p;
  Seg = gft::Image32::Create(dextrseg);
  for(p = 0; p < dextrseg->n; p++)
    Seg->data[p] = ROUND(w1*dextrseg->data[p] +
			 (1.0-w1)*membship->data[p]);
  DrawSeeds(Si, Seg, 255);
  DrawSeeds(Se, Seg, 0);
  return Seg;
}


float MeanProbInS(gft::sImage32 *prob,
		  int *S){
  float sum = 0.0;
  int i,p;
  for(i = 1; i <= S[0]; i++){
    p = S[i];
    sum += prob->data[p];
  }
  return sum/S[0];
}



int main(int argc, char **argv){
  gft::sImage32 *prob,*dextrseg,*energy,*label,*tmp;
  gft::sImage32 *membership8,*ori_img,*mask; //*membership_ift8
  gft::sCImage32f *cimg_lab;
  gft::sCImage *cimg;
  gft::sImageGraph *G;
  int *color;
  int *pixels;
  int *P4, *Si, *Se, *S, *Se_border;
  int M[2];
  IFTData *ift_data;
  int i,n,p,niter,w,h;
  double sum, summin;
  float sharpness, diag, Mdextr;
  ROI roi;
  
  if(argc < 4){
    fprintf(stdout,"usage:\n");
    fprintf(stdout,"relaxeddextr <color_image> <dextr_prob> <4points>\n");
    exit(0);
  }
  
  srand(0);
  cimg = ReadAnyCImage(argv[1]);
  prob = ReadAnyImage(argv[2]);
  P4 = Load4Points(argv[3], prob);
  dextrseg = gft::Image32::Threshold(prob, ROUND(ceilf(0.91*255)), 255);
  //gft::Image32::Write(dextrseg, (char*)"dextrseg.pgm");

  Se = LoadExternalSeeds(dextrseg, P4, 3, prob->ncols+prob->nrows);
  Se_border = LoadExternalSeeds(dextrseg, P4, 3, 5);
  
  cimg_lab = gft::CImage32f::RGB2Lab(cimg);

  energy = ComputeEnergyMap(cimg_lab, prob, dextrseg, Se_border, 0.2, 0.3);

  color = SetColorArray(4);
  ift_data = (IFTData *)malloc(sizeof(IFTData)*(4+1));
  
  for(i = 1; i <= 4; i++){
    M[0] = 1;
    M[1] = P4[i];
    ift_data[i].pred = gft::Image32::Create(energy);
    ift_data[i].cost = gft::AllocDoubleArray(energy->n);

    IFTesum_pow(energy, M,
		ift_data[i].pred,
		ift_data[i].cost, CONNPOWER, NULL);
  }

  pixels = GetCandidatePixels(energy, NULL);
  FourPointConnection(energy, ift_data,
		      color, pixels, false);
  free(pixels);

  Si = LoadConnInternalSeeds(P4);
  S = MergeSeeds(Si, Se);
  //WriteSeeds(Si, Se_border, energy);

  membership8 = gft::Features::KNNFuzzyClassification(cimg_lab,
						      Si, Se_border,
						      290.0*0.03,
						      7, 200, 255);

  mask = gft::Image32::Create(membership8);
  gft::Image32::Set(mask, 1);
  DrawSeeds(S, mask, 0);
  sharpness = gft::Image32::Sharpness(membership8, mask, 255);
  sharpness = sqrtf(sharpness);
  gft::Image32::Destroy(&mask);

  Mdextr = MeanProbInS(dextrseg, Si);
  
  ori_img = SetInitialSeg(prob, membership8,
			  (1.0-sharpness)*powf(Mdextr, 3.0),
			  Si, Se);
  
  for(p = 0; p < dextrseg->n; p++)
    dextrseg->data[p] *= 255;

  label = SetInitialSeg(dextrseg, membership8,
			(1.0-sharpness)*powf(Mdextr, 3.0),
			Si, Se);
  
  G = ComputeGraph(cimg_lab, membership8, sharpness);
  gft::ImageGraph::Orient2Digraph(G, ori_img, 50.0);

  GetROI(&roi, P4, label, 3);
  w = roi.x_max-roi.x_min+1;
  h = roi.y_max-roi.y_min+1;
  diag = sqrtf(w*w + h*h);

  niter = ROUND(diag*0.04*(1.0+1.0*sharpness));  

  gft::ift::ORelax_1(G, S, label, niter);

  gft::Image32::Write(label, (char *)"result.pgm");

  gft::ImageGraph::Destroy(&G);
  gft::CImage32f::Destroy(&cimg_lab);
  gft::Image32::Destroy(&membership8);
  gft::Image32::Destroy(&ori_img);
  gft::Image32::Destroy(&prob);
  gft::Image32::Destroy(&energy);
  gft::Image32::Destroy(&dextrseg);
  gft::Image32::Destroy(&label);
  gft::CImage::Destroy(&cimg);
  for(i = 1; i <= 4; i++){
    gft::Image32::Destroy(&(ift_data[i].pred));
    gft::FreeDoubleArray(&(ift_data[i].cost));
  }
  free(ift_data);
  free(color);
  free(P4);
  free(Si);
  free(Se);
  free(Se_border);
  free(S);
  return 0;
}

