
#include "gft.h"
#include "gft_connseeds.h"



gft::sCImage *ReadAnyCImage(char *file){
  gft::sCImage *cimg;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "ppm") == 0){
    cimg = gft::CImage::Read(file);
  }
  else{
    sprintf(command, "convert %s cimage_tmp.ppm", file);
    system(command);
    cimg = gft::CImage::Read("cimage_tmp.ppm");
    system("rm cimage_tmp.ppm");
  }
  return cimg;
}


gft::sImage32 *ReadAnyImage(char *file){
  gft::sImage32 *img;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "pgm") == 0){
    img = gft::Image32::Read(file);
  }
  else{
    sprintf(command, "convert %s image_tmp.pgm", file);
    system(command);
    img = gft::Image32::Read((char *)"image_tmp.pgm");
    system("rm image_tmp.pgm");
  }
  return img;
}



int main(int argc, char **argv){
  gft::sImage32 *gt, *bin_dextr, *bin, *prob;
  gft::sFileList *Li, *Lg, *L4, *Lp;
  int nimages,i,method=1;
  char files_dir[128];
  char img_file[128];
  char command[512];
  float dice_dextr, jacc_dextr, gbjacc_dextr;
  float dice, jacc, gbjacc;
  float err, err_dextr;
  float sum_dice_dextr = 0.0, sum_jacc_dextr = 0.0, sum_gbjacc_dextr = 0.0;
  float sum_dice = 0.0, sum_jacc = 0.0, sum_gbjacc = 0.0;
  float sum_err = 0.0, sum_err_dextr = 0.0;
  FILE *fp;
  int *P4;
  ROI roi;
  
  if(argc < 3){
    fprintf(stdout,"usage:\n");
    fprintf(stdout,"runexp <database_dir> <method>\n");
    fprintf(stdout,"method: 1......Relaxed DEXTR\n");
    fprintf(stdout,"        2......Grabcut\n");
    exit(0);
  }

  Li = gft::FileList::Create(120);
  sprintf(files_dir, "%s/data", argv[1]);
  gft::FileList::AddFilesInDir(Li, files_dir);

  Lg = gft::FileList::Create(120);
  sprintf(files_dir, "%s/groundtruth", argv[1]);
  gft::FileList::AddFilesInDir(Lg, files_dir);

  L4 = gft::FileList::Create(120);
  sprintf(files_dir, "%s/4points", argv[1]);
  gft::FileList::AddFilesInDir(L4, files_dir);  

  Lp = gft::FileList::Create(120);
  sprintf(files_dir, "%s/dextr_prob", argv[1]);
  gft::FileList::AddFilesInDir(Lp, files_dir);
  
  method = atoi(argv[2]);
  printf("method: %d\n", method);
  nimages = Li->n;
  
  fp = fopen("report.txt", "w");
  
  //------- run exp ----------------------
  fprintf(fp, "img; Dice; Jaccard; G.B.Jacc.; Error\n");

  for(i = 0; i < nimages; i++){
    strcpy(img_file, gft::FileList::GetFile(Li, i));
    gft::FileList::RemoveFileDirectory(img_file);
    
    if(method == 1){
      sprintf(command,"./relaxeddextr %s %s %s",
	      gft::FileList::GetFile(Li, i),
	      gft::FileList::GetFile(Lp, i),
	      gft::FileList::GetFile(L4, i));
      printf("%s\n", command);
      system(command);
    }
    else if(method == 2){
      sprintf(command,"python3 grabcut.py %s %s",
	      gft::FileList::GetFile(Li, i),
	      gft::FileList::GetFile(L4, i));
      printf("%s\n", command);
      system(command);
    }
    
    gt = ReadAnyImage(gft::FileList::GetFile(Lg, i));
    bin = ReadAnyImage((char*)"result.pgm");
    P4 = Load4Points(gft::FileList::GetFile(L4, i), gt);
    GetROI(&roi, P4, gt, 10);
    free(P4);
    
    prob = ReadAnyImage(gft::FileList::GetFile(Lp, i));
    bin_dextr = gft::Image32::Threshold(prob, ROUND(ceilf(0.91*255)), 255);
    
    dice_dextr = gft::Image32::DiceSimilarity(gt, bin_dextr, 128);
    jacc_dextr = gft::Image32::JaccardSimilarity(gt, bin_dextr, 128);
    gbjacc_dextr = gft::Image32::GeneralBalancedJaccard(gt, bin_dextr, 128);
    err_dextr = gft::Image32::ErrorRateWithinBoundingBox(gt, bin_dextr, 128,
							 roi.x_min, roi.y_min,
							 roi.x_max, roi.y_max);
    sum_dice_dextr += dice_dextr;
    sum_jacc_dextr += jacc_dextr;
    sum_gbjacc_dextr += gbjacc_dextr;
    sum_err_dextr += err_dextr;
    
    dice = gft::Image32::DiceSimilarity(gt, bin, 128);
    jacc = gft::Image32::JaccardSimilarity(gt, bin, 128);
    gbjacc = gft::Image32::GeneralBalancedJaccard(gt, bin, 128);
    err = gft::Image32::ErrorRateWithinBoundingBox(gt, bin, 128,
						   roi.x_min, roi.y_min,
						   roi.x_max, roi.y_max);
    fprintf(fp, "%s; %f; %f; %f; %f\n", img_file, dice, jacc, gbjacc, err);
    printf(     "%s; %f; %f; %f; %f\n", img_file, dice, jacc, gbjacc, err);
    
    sum_dice += dice;
    sum_jacc += jacc;
    sum_gbjacc += gbjacc;
    sum_err += err;
    
    gft::Image32::Destroy(&prob);
    gft::Image32::Destroy(&gt);
    gft::Image32::Destroy(&bin);
    gft::Image32::Destroy(&bin_dextr);
  }

  if(method == 1)
    printf("Relaxed Deep Extreme Cut with Connected Extreme Points:\n");
  else
    printf("Grabcut:\n");
  printf("\tdice: %f, jacc: %f, gbjacc: %f, error: %f\n",
	 sum_dice/nimages, sum_jacc/nimages,
	 sum_gbjacc/nimages, sum_err/nimages);
  printf("DEXTR:\n");
  printf("\tdice: %f, jacc: %f, gbjacc: %f, error: %f\n",
	 sum_dice_dextr/nimages, sum_jacc_dextr/nimages,
	 sum_gbjacc_dextr/nimages, sum_err_dextr/nimages);
  
  fclose(fp);
  
  gft::FileList::Destroy(&Li);
  gft::FileList::Destroy(&Lg);
  gft::FileList::Destroy(&L4);
  gft::FileList::Destroy(&Lp);
  return 0;
}

