
#include "gft_connseeds.h"

int *Load4Points(char *filename, gft::sImage32 *img){
  int x,y,p,i;
  FILE *fp;
  int *S;
  S = (int *)malloc(5*sizeof(int));
  fp = fopen(filename, "r");
  S[0] = 4;
  for(i = 1; i <= 4; i++){
    fscanf(fp, "%d %d", &x,&y);
    p = x + y*img->ncols;
    S[i] = p;
  }  
  fclose(fp);
  return S;
}



void GetROI(ROI *roi,
	    int *P4,
	    gft::sImage32 *img,
	    int padding){
  int i,p,x,y;
  roi->y_min = INT_MAX;
  roi->x_min = INT_MAX;
  roi->y_max = INT_MIN;
  roi->x_max = INT_MIN;  
  for(i = 1; i <= P4[0]; i++){
    p = P4[i];
    x = p % img->ncols;
    y = p / img->ncols;
    if(y < roi->y_min) roi->y_min = y;
    if(x < roi->x_min) roi->x_min = x;
    if(y > roi->y_max) roi->y_max = y;
    if(x > roi->x_max) roi->x_max = x;
  }
  roi->y_min = MAX(0, roi->y_min - padding);
  roi->x_min = MAX(0, roi->x_min - padding);
  roi->y_max = MIN(img->nrows-1, roi->y_max + padding);
  roi->x_max = MIN(img->ncols-1, roi->x_max + padding);
}



void IFTesum_pow(gft::sImage32 *energy,
		 int *S,
		 gft::sImage32 *pred,
		 double *cost,
		 double power,
		 int *pixels){
  gft::sHeap64f *Q=NULL;
  gft::sAdjRel *A;
  gft::Pixel u,v;
  int i,p,q,Emax,k;
  double cst;
  A = gft::AdjRel::Neighborhood_8();
  gft::Image32::Set(pred, NIL);
  if(pixels == NULL){
    for(p = 0; p < energy->n; p++)
      cost[p] = DBL_MAX;
  }
  else{
    //memset(cost, 0, sizeof(double)*energy->n);
    bzero(cost, sizeof(double)*energy->n);
    for(k = 1; k <= pixels[0]; k++){
      p = pixels[k];
      cost[p] = DBL_MAX;
    }
  }
  Emax = gft::Image32::GetMaxVal(energy) + 1;
  Q = gft::Heap64f::Create(energy->n, cost);
  for(i = 1; i <= S[0]; i++){
    p = S[i];
    cost[p] = pow(Emax - energy->data[p], power);
    gft::Heap64f::Insert_MinPolicy(Q, p);
  }
  while(!gft::Heap64f::IsEmpty(Q)){
    gft::Heap64f::Remove_MinPolicy(Q, &p);
    
    u.x = p%energy->ncols;
    u.y = p/energy->ncols;
    for (i=1; i < A->n; i++){
      v.x = u.x + A->dx[i];
      v.y = u.y + A->dy[i];
      if (gft::Image32::IsValidPixel(energy, v.x, v.y)){
	q = v.x + v.y*energy->ncols;
	
	cst = cost[p] + pow(Emax - energy->data[q], power);
	
	if(cst < cost[q]){	    
	  gft::Heap64f::Update_MinPolicy(Q, q, cst);
	  pred->data[q] = p;
	}
      }
    }
  }
  gft::AdjRel::Destroy(&A);
  gft::Heap64f::Destroy(&Q);
}


int FindMinJunctionPoint(gft::sImage32 *energy,
			 int Emax,
			 IFTData *ift_data,
			 int n,
			 double *emin,
			 int *pixels){
  int pmin,p,k,i;
  double sum, summin;
  summin = DBL_MAX;
  pmin = 0;

  for(k = 1; k <= pixels[0]; k++){
    p = pixels[k];

    sum = 0.0;
    for(i = 1; i <= n; i++)
      sum += ift_data[i].cost[p];
    sum -= (n-1)*pow(Emax - energy->data[p], CONNPOWER);
    
    if(sum < summin){
      summin = sum;
      pmin = p;
    }
  }
  *emin = summin;
  return pmin;
}



int FindMinJunctionPoint(gft::sImage32 *energy,
			 int Emax,
			 double *cost1,
			 double *cost2,
			 double *cost3,
			 double *emin,
			 int *pixels){
  //gft::sImage32 *img, *roi;
  //gft::Pixel l,h;
  int pmin,p,k;
  double sum, summin;
  summin = DBL_MAX;
  pmin = 0;

  //img = gft::Image32::Create(energy);

  for(k = 1; k <= pixels[0]; k++){
    p = pixels[k];
    sum =  cost1[p];
    sum += cost2[p];
    sum += cost3[p];
    sum -= 2*pow(Emax - energy->data[p], CONNPOWER);
    
    //img->data[p] = ROUND(sum); //ROUND(pow(sum, 1.0/CONNPOWER));
    /*
    if(ROUND(sum) >= 12889 && ROUND(sum) <= 12889 + 50)
      img->data[p] = ROUND(sum) - 12889 + 1;
    else
      img->data[p] = 51;
    */
    
    if(sum < summin){
      summin = sum;
      pmin = p;
    }
  }
  *emin = summin;

  /*
  l.x = pmin % img->ncols - 2;
  l.y = pmin / img->ncols - 7;
  h.x = pmin % img->ncols + 7;
  h.y = pmin / img->ncols + 1;
  roi = gft::Image32::Clone(img, l, h);
  gft::Image32::Write(img, (char *)"conncost.pgm");
  gft::Image32::Write(roi, (char *)"conncost_roi.pgm");  
  gft::Image32::Destroy(&img);
  gft::Image32::Destroy(&roi);  
  */

  return pmin;
}


void ThreePointConnection(gft::sImage32 *energy,
			  IFTData *ift_data,
			  int *color,
			  int *pixels){
  gft::sCImage *cimg;
  int pmin,p,i,Emax;
  double emin;
  Emax = gft::Image32::GetMaxVal(energy) + 1;
  pmin = FindMinJunctionPoint(energy, Emax,
			      ift_data[1].cost,
			      ift_data[2].cost,
			      ift_data[3].cost,
			      &emin,
			      pixels);
  //printf("min: (x,y) = (%d,%d)\n",pmin%energy->ncols, pmin/energy->ncols);
  printf("emin = %lf\n", emin);
  
  cimg = gft::CImage::Clone(energy);
  for(i = 1; i <= 3; i++){  
    p = pmin;
    while(p != NIL){
      cimg->C[0]->data[p] = gft::Color::Channel0(color[i]);
      cimg->C[1]->data[p] = gft::Color::Channel1(color[i]);
      cimg->C[2]->data[p] = gft::Color::Channel2(color[i]);
      p = ift_data[i].pred->data[p];
    }
  }

  cimg->C[0]->data[pmin] = 0;
  cimg->C[1]->data[pmin] = 80;
  cimg->C[2]->data[pmin] = 0;

  gft::CImage::Write(cimg, (char *)"tree_min.ppm");

  gft::CImage::Destroy(&cimg);
}



void WriteFourPointConnection(int p, int q,
			      gft::sImage32 *energy,
			      gft::sImage32 *pred[6],
			      int *color){
  gft::sCImage *cimg;
  int i,t;
  cimg = gft::CImage::Clone(energy);
  for(i = 1; i <= 5; i++){
    if(i == 1 || i == 2)
      t = p;
    else
      t = q;
    while(t != NIL){
      cimg->C[0]->data[t] = gft::Color::Channel0(color[i]);
      cimg->C[1]->data[t] = gft::Color::Channel1(color[i]);
      cimg->C[2]->data[t] = gft::Color::Channel2(color[i]);
      t = pred[i]->data[t];
    }
  }
  
  cimg->C[0]->data[p] = 0;
  cimg->C[1]->data[p] = 80;
  cimg->C[2]->data[p] = 0;

  cimg->C[0]->data[q] = 0;
  cimg->C[1]->data[q] = 80;
  cimg->C[2]->data[q] = 0;
  
  gft::CImage::Write(cimg, (char *)"tree.ppm");
  gft::CImage::Destroy(&cimg);  
}


double OrderedFourPointConnection(gft::sImage32 *energy,
				  IFTData *ift_data,
				  int *color,
				  int *pixels){
  int S[2];
  int p,pmin,q,qmin,Emax,i,k;
  double sum, summin;
  gft::sImage32 *preds[6];
  gft::sImage32 *pred;
  double *cost;
  Emax = gft::Image32::GetMaxVal(energy) + 1;
  pred = gft::Image32::Create(energy);
  cost = gft::AllocDoubleArray(energy->n);
  summin = DBL_MAX;
  pmin = 0;
  qmin = 0;
  for(k = 1; k <= pixels[0]; k++){
    printf("%.2f%%\n",((float)k/(float)pixels[0])*100.0);

    p = pixels[k];
    S[0] = 1;
    S[1] = p;
    IFTesum_pow(energy,	S,
		pred, cost, CONNPOWER, pixels);
    
    q = FindMinJunctionPoint(energy, Emax,
			     cost,
			     ift_data[3].cost,
			     ift_data[4].cost, &sum,
			     pixels);
    sum += ift_data[1].cost[p];
    sum += ift_data[2].cost[p];
    sum -= 2*pow(Emax - energy->data[p], CONNPOWER);

    if(sum < summin){
      summin = sum;
      pmin = p;
      qmin = q;
    }
  }

  printf("emin = %lf\n", summin);

  S[0] = 1;
  S[1] = pmin;
  IFTesum_pow(energy, S,
	      pred, cost, CONNPOWER, pixels);

  preds[0] = NULL;
  preds[1] = ift_data[1].pred;
  preds[2] = ift_data[2].pred;
  preds[3] = ift_data[3].pred;
  preds[4] = ift_data[4].pred;
  preds[5] = pred;
  WriteFourPointConnection(pmin, qmin, energy, preds, color);

  gft::Image32::Destroy(&pred);
  gft::FreeDoubleArray(&cost);

  return summin;
}




double ApproximateOrderedFourPointConnection(gft::sImage32 *energy,
					     IFTData *ift_data,
					     int *color,
					     int *pixels){
  int C,p,q,t,d,i,Emax,ntimes = 5;
  double sum, summin = DBL_MAX;
  gft::sImage32 *preds[6];
  IFTData data;
  int S[2];
  data.pred = gft::Image32::Create(energy);
  data.cost = gft::AllocDoubleArray(energy->n);
  Emax = gft::Image32::GetMaxVal(energy) + 1;
  C = FindMinJunctionPoint(energy, Emax,
			   ift_data,
			   4,
			   &sum,
			   pixels);
  p = C;
  i = 0;
  while(i < ntimes){
    S[0] = 1;
    S[1] = p;
    if(i % 2 == 0) d = 2;
    else           d = 0;
    
    IFTesum_pow(energy, S,
		data.pred,
		data.cost, CONNPOWER, pixels);
    q = FindMinJunctionPoint(energy, Emax,
			     ift_data[1+d].cost,
			     ift_data[2+d].cost,
			     data.cost,
			     &sum,
			     pixels);
    sum += ift_data[1+2-d].cost[p];
    sum += ift_data[2+2-d].cost[p];
    sum -= 2*pow(Emax - energy->data[p], CONNPOWER);

    //printf("sum: %f\n", sum);
    
    if(sum < summin)
      summin = sum;

    t = p; p = q; q = t;
    i++;
  }
  t = p; p = q; q = t;

  preds[0] = NULL;
  preds[1] = ift_data[1+2-d].pred;
  preds[2] = ift_data[2+2-d].pred;
  preds[3] = ift_data[1+d].pred;
  preds[4] = ift_data[2+d].pred;
  preds[5] = data.pred;
  WriteFourPointConnection(p, q, energy, preds, color);
  
  gft::Image32::Destroy(&(data.pred));
  gft::FreeDoubleArray(&(data.cost));

  return sum;
}


void FourPointConnection(gft::sImage32 *energy,
			 IFTData *ift_data,
			 int *color,
			 int *pixels,
			 bool opt){
  IFTData tmp;
  double e, emin;
  char command[512];
  int k;

  emin = DBL_MAX;
  for(k = 2; k <= 4; k++){
    tmp = ift_data[2];
    ift_data[2] = ift_data[k];
    ift_data[k] = tmp;

    if(opt)
      e = OrderedFourPointConnection(energy,
				     ift_data,
				     color,
				     pixels);
    else
      e = ApproximateOrderedFourPointConnection(energy,
						ift_data,
						color,
						pixels);
    sprintf(command, "mv tree.ppm tree_%d.ppm",k-1);
    system(command);
    tmp = ift_data[2];
    ift_data[2] = ift_data[k];
    ift_data[k] = tmp;

    if(e < emin){
      sprintf(command, "cp tree_%d.ppm tree_min.ppm",k-1);
      system(command);
      emin = e;
    }
  }
}



int *SetColorArray(int n){
  int *color;
  int i;
  color = (int *)malloc(sizeof(int)*(n+1));
  color[0] = n;
  color[1] = gft::Color::Triplet(255,   0,   0);
  color[2] = gft::Color::Triplet(  0, 255,   0);
  color[3] = gft::Color::Triplet(  0,   0, 255);
  color[4] = gft::Color::Triplet(255, 255,   0);
  color[5] = gft::Color::Triplet(128,   0, 128);
  for(i = 6; i <= n; i++)
    color[i] = gft::Color::RandomColor();
  return color;
}


int *GetCandidatePixels(gft::sImage32 *energy,
			ROI *roi){
  int *Set = NULL;
  int p,i,j,n,k,Emax;
  ROI Img;
  Img.y_min = 0;
  Img.x_min = 0;
  Img.y_max = energy->nrows-1;
  Img.x_max = energy->ncols-1;
  if(roi == NULL)
    roi = &Img;
  Emax = gft::Image32::GetMaxVal(energy) + 1;
  n = 0;
  for(i = roi->y_min; i <= roi->y_max; i++){
    for(j = roi->x_min; j <= roi->x_max; j++){
      p = j + i*energy->ncols;
      if(energy->data[p] > ROUND(0.1*Emax))
	n++;
    }
  }
  //printf("ratio: %f\n",(float)n/(float)energy->n);  
  Set = (int *)malloc(sizeof(int)*(n+1));
  Set[0] = n;
  k = 1;
  for(i = roi->y_min; i <= roi->y_max; i++){
    for(j = roi->x_min; j <= roi->x_max; j++){
      p = j + i*energy->ncols;
      if(energy->data[p] > ROUND(0.1*Emax)){
	Set[k] = p;
	k++;
      }
    }
  }
  return Set;
}




